from git import Repo
from unidiff import PatchSet
from version_differ.version_differ import get_commit_of_release
import tempfile
from os.path import join, relpath
import os
from package_locator.directory import locate_subdir
from depdive.common import LineDelta, process_whitespace
from collections import defaultdict


class UncertainSubdir(Exception):
    """Cannot verify package directory at version commit"""

    pass


class ReleaseCommitNotFound(Exception):
    pass


class GitError(Exception):
    pass


class FileReadError(Exception):
    def __init__(self, filepath):
        self.file = filepath

    def message(self):
        return "read error for {}".format(self.file)


class SingleCommitFileChangeData:
    def __init__(self, file=None):
        self.source_file: str = file
        self.target_file: str = file
        self.is_rename: bool = False
        self.changed_lines: dict[str, LineDelta] = {}


class MultipleCommitFileChangeData:
    def __init__(self, filename):
        self.filename: str = filename

        # keeps track if it is a renamed file
        self.is_rename: bool = False
        self.old_name: str = None

        self.commits = set()
        self.changed_lines: dict[str, dict[str, LineDelta]] = {}


def get_doubledot_inbetween_commits(repo_path, commit_a, commit_b=""):
    repo = Repo(repo_path)
    commits = repo.iter_commits("{}..{}".format(commit_a, commit_b))
    return [str(c) for c in commits]


def get_all_commits_on_file(repo_path, filepath, start_commit=None, end_commit=None):
    # upto given commit
    repo = Repo(repo_path)

    if start_commit and end_commit:
        commits = repo.git.log(
            "{}^..{}".format(start_commit, end_commit), "--pretty=%H", "--follow", "--", filepath
        ).split("\n")
    elif start_commit:
        commits = repo.git.log("{}^..".format(start_commit), "--pretty=%H", "--follow", "--", filepath).split("\n")
    elif end_commit:
        commits = repo.git.log(end_commit, "--pretty=%H", "--follow", "--", filepath).split("\n")
    else:
        commits = repo.git.log("--pretty=%H", "--follow", "--", filepath).split("\n")

    return list(dict.fromkeys([c for c in commits if c]))


def get_commit_diff(repo_path, commit, reverse=False):
    """
    we do not use git show to get diffs from merge commit
    """
    repo = Repo(repo_path)
    try:
        if not reverse:
            uni_diff_text = repo.git.diff(
                "{}~".format(commit),
                "{}".format(commit),
                "--submodule=diff",
                ignore_blank_lines=True,
                ignore_space_at_eol=True,
            )
        else:
            uni_diff_text = repo.git.diff(
                "{}".format(commit),
                "{}~".format(commit),
                "--submodule=diff",
                ignore_blank_lines=True,
                ignore_space_at_eol=True,
            )
    except:
        # Case 1: first commit, no parent
        uni_diff_text = repo.git.show(
            "{}".format(commit), "--submodule=diff", ignore_blank_lines=True, ignore_space_at_eol=True
        )

    return uni_diff_text


def get_commit_diff_for_file(repo_path, filepath, commit, reverse=False):
    """
    we do not use git show to get diffs from merge commit
    """
    repo = Repo(repo_path)
    try:
        if not reverse:
            uni_diff_text = repo.git.diff(
                "{}~".format(commit),
                "{}".format(commit),
                "--submodule=diff",
                "--",
                filepath,
                ignore_blank_lines=True,
                ignore_space_at_eol=True,
            )
        else:
            uni_diff_text = repo.git.diff(
                "{}".format(commit),
                "{}~".format(commit),
                "--submodule=diff",
                "--",
                filepath,
                ignore_blank_lines=True,
                ignore_space_at_eol=True,
            )
    except:
        # in case of first commit, no parent
        uni_diff_text = repo.git.show(
            "{}".format(commit), "--submodule=diff", "--", filepath, ignore_blank_lines=True, ignore_space_at_eol=True
        )

    return uni_diff_text


def get_inbetween_commit_diff(repo_path, commit_a, commit_b):
    repo = Repo(repo_path)
    uni_diff_text = repo.git.diff(
        "{}".format(commit_a),
        "{}".format(commit_b),
        "--submodule=diff",
        ignore_blank_lines=True,
        ignore_space_at_eol=True,
    )
    return uni_diff_text


def get_inbetween_commit_diff_for_file(repo_path, filepath, commit_a, commit_b):
    repo = Repo(repo_path)
    uni_diff_text = repo.git.diff(
        "{}".format(commit_a),
        "{}".format(commit_b),
        "--submodule=diff",
        "--",
        filepath,
        ignore_blank_lines=True,
        ignore_space_at_eol=True,
    )
    return uni_diff_text


def get_repository_file_list(repo_path, commit):
    repo = Repo(repo_path)
    head = repo.head.object.hexsha

    repo.git.checkout(commit, force=True)
    filelist = []
    for root, dirs, files in os.walk(repo_path):
        for file in files:
            filelist.append(relpath(join(root, file), repo_path))

    repo.git.checkout(head, force=True)
    return set(filelist)


def is_same_commit(sha_a, sha_b):
    return sha_a.startswith(sha_b) or sha_b.startswith(sha_a)


def get_common_ancestor(repo_path, start_commit, end_commit):
    try:
        cmd = "cd {path};git rev-parse $(git log --pretty=%H {start_commit}..{end_commit} | tail -1)^".format(
            path=repo_path, start_commit=start_commit, end_commit=end_commit
        )
        with os.popen(cmd) as process:
            lines = process.readlines()

        assert len(lines) == 1
        ca = lines[0].strip()
        assert ca.isalnum()

        return ca
    except:
        raise GitError


def valid_commit(repo_path, commit):
    repo = Repo(repo_path)
    try:
        repo.commit(commit)
        return True
    except:
        return False


def sort_commits_by_commit_date(repo_path, commits):
    repo = Repo(repo_path)
    sorted_commits = []
    for c in commits:
        c = repo.commit(c)
        d = c.committed_date
        sorted_commits.append((c, d))
    sorted_commits = sorted(sorted_commits, key=lambda x: x[1])
    sorted_commits = [x[0].hexsha for x in sorted_commits]
    return sorted_commits


def get_file_add_commit(repo_path, filepath):
    repo = Repo(repo_path)
    commits = repo.git.log("--pretty=%H", "--diff-filter=A", "--", filepath).split("\n")
    return commits[0]


class RepositoryDiff:
    def __init__(
        self, ecosystem, package, repository, old_version, new_version, old_version_commit=None, new_version_commit=None
    ):
        self.ecosystem = ecosystem
        self.package = package
        self.repository = repository
        self.old_version = old_version
        self.new_version = new_version

        self._temp_dir = None
        self.repo_path = None

        self.old_version_commit = old_version_commit
        self.new_version_commit = new_version_commit

        self.old_version_subdir = None  # package directory at the old version commit
        self.new_version_subdir = None  # package directory at the new version commit
        self.common_ancestor_commit_new_and_old_version = None

        self.submodule_paths = []

        self.commits = None
        self.reverse_commits = None

        self.diff = None  # diff across individual commits
        self.new_version_filelist = None
        self.single_diff = None  # single diff from old to new

        self.build_repository_diff()

    def get_commit_of_release(self, version):
        repo = Repo(self.repo_path)
        tags = repo.tags
        c = get_commit_of_release(tags, self.package, version)
        if c:
            return c.hexsha

    def cleanup(self):
        self._temp_dir.cleanup()

    def _process_submodules(self):
        repo = Repo(self.repo_path)
        repo.submodule_update(recursive=True, init=True)

        head = repo.head.object.hexsha
        try:
            repo.git.checkout(self.old_version_commit, force=True)
            repo.submodule_update(recursive=True)
            self.submodule_paths += [x.path for x in repo.submodules]
        except:
            pass
        try:
            repo.git.checkout(self.new_version_commit, force=True)
            repo.submodule_update(recursive=True)
            self.submodule_paths += [x.path for x in repo.submodules]
        except:
            pass

        repo.git.checkout(head, force=True)

        self.submodule_paths = list(set(self.submodule_paths))

    def build_repository_diff(self):
        if not self.repo_path:
            self._temp_dir = tempfile.TemporaryDirectory()
            self.repo_path = self._temp_dir.name
            Repo.clone_from(self.repository, self.repo_path)

        if (
            not self.old_version_commit
            or not self.new_version_commit
            or not valid_commit(self.repo_path, self.old_version_commit)
            or not valid_commit(self.repo_path, self.new_version_commit)
        ):
            self.old_version_commit = self.get_commit_of_release(self.old_version)
            self.new_version_commit = self.get_commit_of_release(self.new_version)
            if not self.old_version_commit or not self.new_version_commit:
                raise ReleaseCommitNotFound

        try:
            self.old_version_subdir = locate_subdir(
                self.ecosystem, self.package, self.repository, commit=self.old_version_commit, version=self.old_version
            )
            self.new_version_subdir = locate_subdir(
                self.ecosystem, self.package, self.repository, commit=self.new_version_commit, version=self.new_version
            )
        except:
            self._temp_dir.cleanup()
            raise UncertainSubdir

        self.common_ancestor_commit_new_and_old_version = get_common_ancestor(
            self.repo_path, self.old_version_commit, self.new_version_commit
        )

        self._process_submodules()

        self.commits = set(
            get_doubledot_inbetween_commits(self.repo_path, self.old_version_commit, self.new_version_commit)
        )
        self.reverse_commits = set(
            get_doubledot_inbetween_commits(self.repo_path, self.new_version_commit, self.old_version_commit)
        )

        self.new_version_filelist = get_repository_file_list(self.repo_path, self.new_version_commit)

        self.diff = self.get_commit_diff_stats_from_repo(self.repo_path, list(self.commits), list(self.reverse_commits))

        self.single_diff = self.get_diff_files(
            get_inbetween_commit_diff(self.repo_path, self.old_version_commit, self.new_version_commit)
        )

    def get_full_file_single_diff(self, filepath):
        single_diff = SingleCommitFileChangeData(filepath)
        try:
            with open(join(self.repo_path, filepath), "r") as f:
                lines = f.readlines()
        except:
            raise FileReadError(filepath)
        for l in lines:
            l = process_whitespace(l)
            if l:
                single_diff.changed_lines[l] = single_diff.changed_lines.get(l, LineDelta())
                single_diff.changed_lines[l].additions += 1
        return single_diff

    def get_full_file_history(self, filepath, end_commit="HEAD"):
        """get commit history of filepath upto given commit point"""
        commits = get_all_commits_on_file(self.repo_path, filepath, end_commit=end_commit)
        if filepath in self.diff and not set(commits) - self.diff[filepath].commits:
            return

        add_commit = get_file_add_commit(self.repo_path, filepath)
        if filepath in self.diff and add_commit in self.diff[filepath].commits:
            return

        single_diff = self.get_full_file_single_diff(filepath)

        diff = self.get_commit_diff_stats_from_repo(self.repo_path, commits)
        if filepath in diff:
            self.diff[filepath] = self.diff.get(filepath, MultipleCommitFileChangeData(filepath))
            for line in diff[filepath].changed_lines.keys():
                self.diff[filepath].changed_lines[line] = self.diff[filepath].changed_lines.get(line, {})
                for commit in diff[filepath].changed_lines[line].keys():
                    if commit not in self.diff[filepath].changed_lines[line]:
                        self.diff[filepath].commits.add(commit)
                        self.diff[filepath].changed_lines[line][commit] = diff[filepath].changed_lines[line]
            self.commits |= self.diff[filepath].commits

        self.single_diff[filepath] = single_diff

    def traverse_beyond_new_version_commit(self, filepath, phantom_lines):
        """
        returns new possible commit boundary
        for the corner case where the version was wrongly tagged at some commit
        and the actual uploaded artifact contains one or more commits beyond
        the boundary pointed at by the version tag

        Note that, we expand our commit boundary very conservatively,
        if the immediate next commit outside the boundary on the given file
        does not address phantom lines, we quit.

        We only do it for the new version commit,
        as old version does not affect phantom lines present in the new update.

        Also, if commit boundary has changed, we re-process the object
        """

        new_version_commit, old_version_commit = self.new_version_commit, self.old_version_commit
        if not new_version_commit or not old_version_commit:
            return False

        # check if there's any phantom addition
        additions = 0
        for l in phantom_lines.keys():
            additions += phantom_lines[l].additions
        if additions == 0:
            return False

        # first take a look at the commis afterward new_version_commits
        commits = get_doubledot_inbetween_commits(self.repo_path, new_version_commit)[::-1]
        if commits:
            commits = commits[1:] if commits[0] == new_version_commit else commits
            for commit in commits:
                diff = self.get_diff_files(get_commit_diff(self.repo_path, commit))
                commit_outside_boundary = True  # assume this commit is outside the actual boundary
                if filepath in diff:
                    commit_diff = diff[filepath].changed_lines
                    for line in commit_diff.keys():
                        p_line = process_whitespace(line)
                        if p_line in phantom_lines.keys():
                            phantom_lines[p_line].subtract(commit_diff[line])
                            if phantom_lines[p_line].additions == 0:
                                phantom_lines.pop(p_line)
                                new_version_commit = commit
                                commit_outside_boundary = False
                if commit_outside_boundary or not phantom_lines:
                    break

        if new_version_commit != self.new_version_commit or old_version_commit != self.old_version_commit:
            # get new version commit
            # however, the next commit can be from another branch
            # was merged with cur new version commit afterwards
            # if that's the case we want the merge commit
            after_commits = get_doubledot_inbetween_commits(self.repo_path, new_version_commit)[::-1]
            if self.new_version_commit in after_commits:
                idx = after_commits.index(self.new_version_commit)
                if idx < len(after_commits) - 1:
                    new_version_commit = after_commits[idx + 1]
                else:
                    return False

            # sanity check
            new_inbetween_commits = set(
                get_doubledot_inbetween_commits(self.repo_path, old_version_commit, new_version_commit)
            )
            old_inbetween_commits = set(
                get_doubledot_inbetween_commits(self.repo_path, self.old_version_commit, self.new_version_commit)
            )
            if not old_inbetween_commits - new_inbetween_commits:
                self.new_version_commit = new_version_commit
                self.build_repository_diff()
                return True

        return False

    def get_commit_diff_stats_from_repo(self, repo_path, commits, reverse_commits=[]):
        files = {}
        for commit in commits + reverse_commits:
            diff = self.get_diff_files(get_commit_diff(repo_path, commit, reverse=commit in reverse_commits))
            for file in diff.keys():
                files[file] = files.get(file, MultipleCommitFileChangeData(file))
                if diff[file].is_rename:
                    files[file].is_rename = True
                    files[file].old_name = diff[file].source_file
                for line in diff[file].changed_lines.keys():
                    files[file].changed_lines[line] = files[file].changed_lines.get(line, {})
                    assert commit not in files[file].changed_lines[line]
                    files[file].commits.add(commit)
                    files[file].changed_lines[line][commit] = diff[file].changed_lines[line]

        def recurring_merge_rename(f, merged_files):
            merged_files.add(f)
            if files[f].is_rename and files[f].old_name in files.keys() and files[f].old_name not in merged_files:
                old_f = files[f].old_name
                files[old_f] = recurring_merge_rename(old_f, merged_files)
                for l in files[old_f].changed_lines.keys():
                    if l not in files[f].changed_lines:
                        files[f].changed_lines[l] = files[old_f].changed_lines[l]
                    else:
                        for c in files[old_f].changed_lines[l]:
                            if c not in files[f].changed_lines[l]:
                                files[f].changed_lines[l][c] = files[old_f].changed_lines[l][c]
                files[f].commits |= files[old_f].commits
            return files[f]

        # converge with old name in the case of renamed files
        for f in files.keys():
            files[f] = recurring_merge_rename(f, set())

        def get_all_old_names(f, merged_files):
            merged_files.add(f)
            if f not in files or not files[f].is_rename or files[f].old_name in merged_files:
                return []
            else:
                return [files[f].old_name] + get_all_old_names(files[f].old_name, merged_files)

        rename_map = {}
        all_old_names = set()
        for f in files.keys():
            rename_map[f] = set(get_all_old_names(f, set()))
            all_old_names |= rename_map[f]

        for f in list(files.keys()):
            if f not in all_old_names and rename_map[f]:
                for old_f in rename_map[f]:
                    files[old_f] = files[f]

        return files

    def process_patch_filepath(self, filepath):
        filepath = filepath.removeprefix("a/")
        filepath = filepath.removeprefix("b/")
        if filepath == "/dev/null":
            filepath = None
        return filepath

    def get_diff_files(self, uni_diff_text):
        patch_set = PatchSet(uni_diff_text)
        files = {}

        for patched_file in patch_set:
            f = SingleCommitFileChangeData()
            f.source_file = self.process_patch_filepath(patched_file.source_file)
            f.target_file = self.process_patch_filepath(patched_file.target_file)

            f.is_rename = patched_file.is_rename

            add_lines = [
                process_whitespace(line.value.strip()) for hunk in patched_file for line in hunk if line.is_added
            ]

            del_lines = [
                process_whitespace(line.value.strip()) for hunk in patched_file for line in hunk if line.is_removed
            ]

            for line in del_lines:
                if line:
                    f.changed_lines[line] = f.changed_lines.get(line, LineDelta())
                    f.changed_lines[line].deletions += 1

            for line in add_lines:
                if line:
                    f.changed_lines[line] = f.changed_lines.get(line, LineDelta())
                    f.changed_lines[line].additions += 1

            files[patched_file.path] = f

        return files

    def git_blame_delete(self, filepath, start_commit, new_version_commit):
        try:
            with open(join(self.repo_path, filepath), "r") as f:
                filelines = f.readlines()
        except:
            raise FileReadError(filepath)
        filelines = [process_whitespace(l.strip()) for l in filelines]

        cmd = "cd {path};git blame --reverse -l {start_commit}..{end_commit} {fname}".format(
            path=self.repo_path, start_commit=start_commit, end_commit=new_version_commit, fname=filepath
        )
        with os.popen(cmd) as process:
            blame = process.readlines()

        if not len(blame) == len(filelines):
            raise GitError

        blame = [line.split(" ")[0] for line in blame]
        blame = [line.removeprefix("^") for line in blame]
        blame_map = defaultdict(list)
        for i, c in enumerate(blame):
            blame_map[c] += [i]

        def find_removal_commit(line, candidate_commits):
            """
            git blame may be inaccurate in case of some bulk changes,
            so run a validation step,
            by checking if the commit has indeed deleted the line

            assumption: commits are sorted old to new
            TODO: do explicit sorting within the function, maybe not because inner function
            """
            if line in self.diff[filepath].changed_lines:
                for commit in candidate_commits:
                    if (
                        commit in self.diff[filepath].changed_lines[line]
                        and self.diff[filepath].changed_lines[line][commit].deletions > 0
                    ):
                        return commit

        c2c = defaultdict(list)
        for commit in blame_map.keys():
            if not commit.isalnum():
                raise GitError
            if is_same_commit(commit, new_version_commit):
                continue

            next_commits = get_doubledot_inbetween_commits(self.repo_path, commit, new_version_commit)[::-1]
            next_commits = list(filter(lambda c: c in self.diff[filepath].commits, next_commits))
            blame_map[commit] = list(
                filter(lambda i: filelines[i] in self.diff[filepath].changed_lines, blame_map[commit])
            )
            for i in blame_map[commit]:
                next_commit = find_removal_commit(filelines[i], next_commits)
                if next_commit:
                    c2c[next_commit] += [filelines[i]]
                else:
                    # possible explanations
                    # 1. blank line
                    # 2. deletion commit present in both old_version_commit and new_version_commit
                    # 3. line present in new version, git blame error due to bulk change somewhere
                    pass
        return c2c

    def git_blame(self, filepath, commit):
        c2c = defaultdict(list)  # commit to code
        repo = Repo(self.repo_path)
        for commit, lines in repo.blame(commit, filepath):
            c2c[commit.hexsha] += list(lines)
        return c2c



from version_differ.version_differ import FileDiff
from package_locator.locator import get_repository_url_and_subdir
from depdive.common import LineDelta, process_whitespace
from depdive.registry_diff import get_registry_version_diff
from depdive.repository_diff import (
    RepositoryDiff,
    SingleCommitFileChangeData,
    get_repository_file_list,
    UncertainSubdir,
    sort_commits_by_commit_date,
)
from depdive.code_review_checker import CommitReviewInfo
from git import Repo
import os


class PackageDirectoryChanged(Exception):
    pass


class DepdiveStats:
    def __init__(
        self,
        added_reviewed_lines,
        added_non_reviewed_lines,
        removed_reviewed_lines,
        removed_non_reviewed_lines,
        reviewed_commits,
        non_reviewed_commits,
        phantom_files,
        files_with_phantom_lines,
        phantome_lines,
    ) -> None:
        self.added_reviewed_lines = added_reviewed_lines
        self.added_non_reviewed_lines = added_non_reviewed_lines
        self.removed_reviewed_lines = removed_reviewed_lines
        self.removed_non_reviewed_lines = removed_non_reviewed_lines

        self.reviewed_lines = self.added_reviewed_lines + self.removed_reviewed_lines
        self.non_reviewed_lines = self.added_non_reviewed_lines + self.removed_non_reviewed_lines

        self.total_commit_count = len(reviewed_commits) + len(non_reviewed_commits)
        self.reviewed_commit_count = len(reviewed_commits)

        self.reviewed_commits = reviewed_commits
        self.non_reviewed_commits = non_reviewed_commits

        self.phantom_files = phantom_files
        self.files_with_phantom_lines = files_with_phantom_lines
        self.phantom_lines = phantome_lines

    def print(self):
        print(self.reviewed_commits, self.non_reviewed_commits)
        print(
            self.added_reviewed_lines,
            self.added_non_reviewed_lines,
            self.removed_reviewed_lines,
            self.removed_non_reviewed_lines,
        )
        print(self.phantom_files, self.files_with_phantom_lines, self.phantom_lines)
        print(self.reviewed_lines, self.non_reviewed_lines, self.total_commit_count, self.reviewed_commit_count)


class CodeReviewAnalysis:
    def __init__(self, ecosystem, package, old_version, new_version, repository=None, directory=None):
        self.ecosystem: str = ecosystem
        self.package: str = package
        self.old_version: str = old_version
        self.new_version: str = new_version

        self.repository: str = repository
        self.directory: str = directory
        if not self.repository:
            self._locate_repository()

        self.start_commit: str = None
        self.end_commit: str = None
        self.common_starter_commit: str = None

        # line counter from version_differ output
        self.registry_diff = {}

        # Newly added phantom files:
        # files present in the registry, but not in repo
        self.phantom_files: set[str] = set()

        # files present in repo,
        # but removed in the new version in the registry
        self.removed_files_in_registry: dict[str, FileDiff] = {}

        # Newly added phantom lines:
        # files present in repo but contains lines
        # that are only present in registry
        self.phantom_lines: dict[str, dict[str, LineDelta]] = {}

        # code to commit mapping
        self.added_loc_to_commit_map: dict[str, dict[str, list(str)]] = {}
        self.removed_loc_to_commit_map: dict[str, dict[str, list(str)]] = {}

        # commit to review map
        self.commit_review_info: dict[str, CommitReviewInfo] = {}

        self.stats: DepdiveStats = None

        self.run_analysis()

    def _locate_repository(self):
        self.repository, self.directory = get_repository_url_and_subdir(self.ecosystem, self.package)

    def get_repo_path_from_registry_path(self, filepath, repository_diff):

        subdir = self.directory.removeprefix("./").removesuffix("/")
        repo_f = subdir + "/" + filepath if subdir else filepath

        # TODO: below hack should go away?
        if filepath == "LICENSE":
            if repo_f not in repository_diff.new_version_filelist and filepath in repository_diff.new_version_filelist:
                return filepath

        # check if symlink
        fullpath = os.path.join(repository_diff.repo_path, repo_f)
        if os.path.islink(fullpath):
            target_link = os.path.realpath(os.path.join(fullpath, os.readlink(fullpath)))
            repo_f = os.path.relpath(target_link, os.path.realpath(repository_diff.repo_path))

        return repo_f

    def _process_phantom_files(self, registry_diff, repository_diff):
        """
        Phantom files: Files that are present in the registry,
                        but not in the source repository

        Also, keep track of files that are present in the repository,
        but have been removed in the new version
        """
        for f in registry_diff.new_version_filelist:
            repo_f = self.get_repo_path_from_registry_path(f, repository_diff)
            if repo_f not in repository_diff.new_version_filelist:
                self.phantom_files.add(f)

        for f in registry_diff.diff.keys():
            repo_f = self.get_repo_path_from_registry_path(f, repository_diff)
            if not registry_diff.diff[f].target_file and repo_f in repository_diff.new_version_filelist:
                self.removed_files_in_registry[f] = registry_diff.diff[f]

    def _get_phantom_lines_in_a_file(self, registry_file_diff, repo_file_diff):
        phantom = {}
        # TODO: should we check lines in the repository file that are not in registry file?
        #      the removal can be dangerous as well
        for l in registry_file_diff:
            if (
                l not in repo_file_diff.changed_lines
                or registry_file_diff[l].delta() != repo_file_diff.changed_lines[l].delta()
            ):
                phantom[l] = LineDelta(registry_file_diff[l].additions, registry_file_diff[l].deletions)
                if l in repo_file_diff.changed_lines:
                    phantom[l].subtract(repo_file_diff.changed_lines[l])

        return phantom

    def _get_registry_file_line_counter(self, f):
        lc = {}
        for l in [process_whitespace(l) for l in f.added_lines]:
            lc[l] = lc.get(l, LineDelta())
            lc[l].additions += 1

        for l in [process_whitespace(l) for l in f.removed_lines]:
            lc[l] = lc.get(l, LineDelta())
            lc[l].deletions += 1

        return lc

    def _proccess_phantom_lines(self, registry_diff, repository_diff):
        new_version_repo_filelist = get_repository_file_list(
            repository_diff.repo_path, repository_diff.new_version_commit
        )

        repo = Repo(repository_diff.repo_path)
        head = repo.head.object.hexsha
        repo.git.checkout(repository_diff.new_version_commit, force=True)

        for f in registry_diff.diff.keys():
            registry_file_diff = self._get_registry_file_line_counter(registry_diff.diff[f])
            self.registry_diff[f] = registry_file_diff

            if not registry_diff.diff[f].target_file:
                continue

            repo_f = self.get_repo_path_from_registry_path(f, repository_diff)

            if (
                not registry_diff.diff[f].source_file
                and repo_f in new_version_repo_filelist
                and (repo_f not in repository_diff.diff.keys() or not repository_diff.diff[repo_f].is_rename)
            ):
                # possible explanation: newly included file to be published - get all the commits from the beginning
                # get full file history for such files
                repository_diff.get_full_file_history(repo_f, end_commit=repository_diff.new_version_commit)

            if (
                not registry_diff.diff[f].source_file
                and not registry_diff.diff[f].is_rename
                and repo_f in repository_diff.diff.keys()
                and repository_diff.diff[repo_f].is_rename
            ):
                single_diff = repository_diff.get_full_file_single_diff(repo_f)
            else:
                single_diff = repository_diff.single_diff.get(repo_f, SingleCommitFileChangeData())

            phantom_lines = self._get_phantom_lines_in_a_file(registry_file_diff, single_diff)
            if phantom_lines:
                # try looking beyond the initial commit boundary
                repo.git.checkout(head, force=True)
                has_commit_boundary_changed = repository_diff.traverse_beyond_new_version_commit(
                    repo_f,
                    phantom_lines.copy(),
                )
                if has_commit_boundary_changed:
                    return False
                repo.git.checkout(repository_diff.new_version_commit, force=True)

            if phantom_lines:
                self.phantom_lines[f] = phantom_lines

        repo.git.checkout(head, force=True)
        return True

    def _filter_out_phantom_files(self, registry_diff):
        for pf in self.phantom_files:
            registry_diff.diff.pop(pf, None)

        for pf in self.removed_files_in_registry.keys():
            registry_diff.diff.pop(pf, None)

        return registry_diff

    def run_analysis(self):
        if not self.repository:
            self._locate_repository()

        registry_diff = get_registry_version_diff(self.ecosystem, self.package, self.old_version, self.new_version)
        repository_diff = RepositoryDiff(
            self.ecosystem,
            self.package,
            self.repository,
            self.old_version,
            self.new_version,
            old_version_commit=registry_diff.old_version_git_sha,
            new_version_commit=registry_diff.new_version_git_sha,
        )

        # checking package directory
        if repository_diff.old_version_subdir != repository_diff.new_version_subdir:
            raise PackageDirectoryChanged
        elif not repository_diff.new_version_subdir:
            raise UncertainSubdir
        if repository_diff.new_version_subdir != self.directory:
            self.directory = repository_diff.new_version_subdir

        self._process_phantom_files(registry_diff, repository_diff)
        self._filter_out_phantom_files(registry_diff)

        phantom_lines_processed = False
        while not phantom_lines_processed:
            phantom_lines_processed = self._proccess_phantom_lines(registry_diff, repository_diff)

        self.start_commit = repository_diff.old_version_commit
        self.end_commit = repository_diff.new_version_commit

        self.map_commit_to_added_lines(repository_diff, registry_diff)
        self.map_commit_to_removed_lines(repository_diff, registry_diff)

        for f in registry_diff.diff.keys():
            repo_files = [self.get_repo_path_from_registry_path(f, repository_diff)]
            if registry_diff.diff[f].is_rename:
                repo_files += [
                    self.get_repo_path_from_registry_path(registry_diff.diff[f].source_file, repository_diff)
                ]
            for repo_f in repo_files:
                if repo_f in repository_diff.diff.keys():
                    for commit in repository_diff.diff[repo_f].commits:
                        if commit not in self.commit_review_info:
                            self.commit_review_info[commit] = CommitReviewInfo(self.repository, commit)

        self.stats = self.get_stats()
        repository_diff.cleanup()

    def map_commit_to_added_lines(self, repository_diff, registry_diff):
        def map_submdule_to_added_lines(f, repo_f):
            for path in repository_diff.submodule_paths:
                if "{}/".format(path) in repo_f:
                    commits = sort_commits_by_commit_date(
                        repository_diff.repo_path, list(repository_diff.diff[repo_f].commits)
                    )
                    assert commits, "no commit found for submodule {}".format(path)
                    commit = commits[-1]

                    added_lines = [process_whitespace(l) for l in registry_diff.diff[f].added_lines]
                    added_lines = [l for l in added_lines if l]
                    self.added_loc_to_commit_map[f] = {commit: added_lines}
                    return True
            return False

        files_with_added_lines = set()
        for f in registry_diff.diff.keys():
            if registry_diff.diff[f].target_file and registry_diff.diff[f].added_lines:
                files_with_added_lines.add(registry_diff.diff[f].target_file)

        for f in files_with_added_lines:
            repo_f = self.get_repo_path_from_registry_path(f, repository_diff)

            # ignore files with only phantom line changes
            if f in self.phantom_lines.keys() and repo_f not in repository_diff.diff.keys():
                continue

            # check if file is in a submodule
            if map_submdule_to_added_lines(f, repo_f):
                continue

            c2c = repository_diff.git_blame(repo_f, repository_diff.new_version_commit)

            for commit in list(c2c.keys()):
                if commit not in repository_diff.diff[repo_f].commits:
                    c2c.pop(commit)
                else:
                    c2c[commit] = [process_whitespace(l) for l in c2c[commit]]
                    c2c[commit] = [l for l in c2c[commit] if l]

            self.added_loc_to_commit_map[f] = c2c

    def map_commit_to_removed_lines(self, repository_diff, registry_diff):
        def map_submdule_to_removed_lines(f, repo_f):
            for path in repository_diff.submodule_paths:
                if "{}/".format(path) in repo_f:
                    commits = sort_commits_by_commit_date(
                        repository_diff.repo_path, list(repository_diff.diff[repo_f].commits)
                    )
                    assert commits, "no commit found for submodule {}".format(path)
                    commit = commits[0]

                    removed_lines = [process_whitespace(l) for l in registry_diff.diff[f].removed_lines]
                    removed_lines = [l for l in removed_lines if l]
                    self.removed_loc_to_commit_map[f] = {commit: removed_lines}
                    return True
            return False

        starter_point_file_list = get_repository_file_list(
            repository_diff.repo_path, repository_diff.common_ancestor_commit_new_and_old_version
        )

        files_with_removed_lines = set()
        for f in registry_diff.diff.keys():
            if registry_diff.diff[f].source_file and registry_diff.diff[f].removed_lines:
                files_with_removed_lines.add(registry_diff.diff[f].source_file)

        repo = Repo(repository_diff.repo_path)
        head = repo.head.object.hexsha
        repo.git.checkout(repository_diff.common_ancestor_commit_new_and_old_version, force=True)

        for f in files_with_removed_lines:
            repo_f = self.get_repo_path_from_registry_path(f, repository_diff)

            # file may not be in version diff in repo
            # possible explanations:
            # 1. file may not be in the common starter point at all
            # 2. phantom line changes
            if repo_f not in starter_point_file_list or repo_f not in repository_diff.diff.keys():
                continue

            if map_submdule_to_removed_lines(f, repo_f):
                continue

            c2c = repository_diff.git_blame_delete(
                repo_f, repository_diff.common_ancestor_commit_new_and_old_version, repository_diff.new_version_commit
            )
            for commit in list(c2c.keys()):
                if commit not in repository_diff.commits:
                    c2c.pop(commit)
                else:
                    c2c[commit] = [process_whitespace(l) for l in c2c[commit]]
                    c2c[commit] = [l for l in c2c[commit] if l]

            self.removed_loc_to_commit_map[f] = c2c
        repo.git.checkout(head, force=True)

    def get_stats(self):
        added_reviewed_lines = added_non_reviewed_lines = 0
        non_reviewed_commits = set()
        reviewed_commits = set()

        for f in self.added_loc_to_commit_map.keys():
            for commit in self.added_loc_to_commit_map[f].keys():
                cur = len(self.added_loc_to_commit_map[f][commit])
                if self.commit_review_info[commit].review_category:
                    added_reviewed_lines += cur
                    reviewed_commits.add(commit)
                else:
                    added_non_reviewed_lines += cur
                    non_reviewed_commits.add(commit)

        removed_reviewed_lines = removed_non_reviewed_lines = 0
        for f in self.removed_loc_to_commit_map.keys():
            for commit in self.removed_loc_to_commit_map[f].keys():
                cur = len(self.removed_loc_to_commit_map[f][commit])
                if self.commit_review_info[commit].review_category:
                    removed_reviewed_lines += cur
                    reviewed_commits.add(commit)
                else:
                    removed_non_reviewed_lines += cur
                    non_reviewed_commits.add(commit)

        phantom_files = len(self.phantom_files)

        files_with_phantom_lines = set()
        phantom_lines = 0
        for f in self.phantom_lines.keys():
            for l in self.phantom_lines[f].keys():
                if self.phantom_lines[f][l].additions > 0:
                    files_with_phantom_lines.add(f)
                    phantom_lines += self.phantom_lines[f][l].additions
        files_with_phantom_lines = len(files_with_phantom_lines)

        return DepdiveStats(
            added_reviewed_lines,
            added_non_reviewed_lines,
            removed_reviewed_lines,
            removed_non_reviewed_lines,
            reviewed_commits,
            non_reviewed_commits,
            phantom_files,
            files_with_phantom_lines,
            phantom_lines,
        )



from package_locator.common import CARGO, PYPI
from version_differ.common import PIP
from version_differ.version_differ import get_version_diff_stats


class VersionDifferError(Exception):
    pass


def get_registry_version_diff(ecosystem, package, old, new):
    if ecosystem == PYPI:
        ecosystem = PIP

    try:
        version_diff = get_version_diff_stats(ecosystem, package, old, new)
    except:
        raise VersionDifferError

    # preprocess auto-gen files respective to each registry
    if ecosystem == CARGO:
        preprocess_cargo_crate_files(version_diff)

    return version_diff


def preprocess_cargo_crate_files(version_diff):
    # filter out auto-generated files
    auto_gen_files = [".cargo_vcs_info.json", "Cargo.lock"]
    for f in auto_gen_files:
        version_diff.diff.pop(f, None)
        version_diff.new_version_filelist.discard(f)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['depdive']

package_data = \
{'': ['*']}

install_requires = \
['PyGithub>=1.55,<2.0',
 'PyYAML>=5.4.1,<6.0.0',
 'click>=8.0.0,<9.0.0',
 'package-locator>=0.4.4,<0.5.0',
 'rich>=10.3.0,<11.0.0',
 'version-differ>=0.3.14,<0.4.0']

entry_points = \
{'console_scripts': ['depdive = depdive.__main__:main']}

setup_kwargs = {
    'name': 'depdive',
    'version': '0.0.41',
    'description': 'Performs security checks for a dependency update',
    'long_description': 'DepDive\n===========================\n|PyPI| |Python Version| |License| |Read the Docs| |Build| |Tests| |Codecov| |pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/depdive.svg\n   :target: https://pypi.org/project/depdive/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/depdive\n   :target: https://pypi.org/project/depdive\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/github/license/nasifimtiazohi/depdive\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/depdive/latest.svg?label=Read%20the%20Docs\n   :target: https://depdive.readthedocs.io/\n   :alt: Read the documentation at https://depdive.readthedocs.io/\n.. |Build| image:: https://github.com/nasifimtiazohi/depdive/workflows/Build%20depdive%20Package/badge.svg\n   :target: https://github.com/nasifimtiazohi/depdive/actions?workflow=Package\n   :alt: Build Package Status\n.. |Tests| image:: https://github.com/nasifimtiazohi/depdive/workflows/Run%20depdive%20Tests/badge.svg\n   :target: https://github.com/nasifimtiazohi/depdive/actions?workflow=Tests\n   :alt: Run Tests Status\n.. |Codecov| image:: https://codecov.io/gh/nasifimtiazohi/depdive/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/nasifimtiazohi/depdive\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n   \n  \n\nEach time you accept a dependency update, you are pulling in new third-party code. However, how to ensure that the update is secure? \n\nOne way to put a security (and quality) control check before accepting an update is to check what part of the code changes have passed through a code review process. DepDive automates this check.\n\n\nWorkflow\n--------\n\nDepDive maps code changes between two versions of a package uploaded in the registry to the corresponding commits that made these changes in the package\'s \nsource code repository. DepDive then identifies if there was a code reviewer for the mapped commits through rule-based checks for evidences of code review on GitHub.\n\nAlong the process, DepDive also outputs phantom artifacts, that is, files and lines that are present in the registry, but not present in the repository. Example of phantom files can be compiles binaries in PyPI packages, transpiled JavaScript in npm, and other auto-generated files. Not to mention, malicious actors can sneak in code in the last mile between the repository and the registry. \n\nDepDive works for four package registries: (1) Crates.io (Rust), (2) npm (JavaScript), (3) PyPI (Python), (4) RubyGems (Ruby).\nCurrently, DepDive only works for GitHub repositories, as GitHub is our primary source to check for code-review. \n\n\n.. image:: docs/images/depdive.drawio.png\n\nFeatures\n--------\n\n* Outputs changes that have been code reviewed and that have not in a depednency update. You can calculate the code reviewe coverage from the output by diving the revewed lines of code changes by the total lines of code of changes in an update.\n* Outputs the reviewed and non-reviewed commits, and how we determined if a commit was code-reviewed, and the actors involved in the review.\n* Outputs files that are present in the update version downloaded from the registry, but not in the repository.\n* Updates lines of code changes that are present in the udpate, but cannot be mapped to changes in the repository.\n\n\nInstallation\n------------\n\nYou can install *depdive* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install depdive\n\n\nUsage\n-----\n.. code-block:: python\n\n    ca = CodeReviewAnalysis(CARGO, "tokio", "1.8.4", "1.9.0")\n    stats = ca.stats\n    assert stats.phantom_files == 0\n    assert stats.files_with_phantom_lines == 0\n    assert stats.phantom_lines == 0\n    assert stats.reviewed_lines == 3694\n    assert stats.non_reviewed_lines == 0\n    assert stats.total_commit_count == 29\n    assert stats.reviewed_commit_count == 29\n    \n    ca = CodeReviewAnalysis(RUBYGEMS, "pundit", "2.1.0", "2.1.1")\n    stats = ca.stats\n    assert stats.phantom_files == 0\n    assert stats.files_with_phantom_lines == 0\n    assert stats.phantom_lines == 0\n    assert stats.reviewed_lines == 128\n    assert stats.non_reviewed_lines == 186\n    assert stats.total_commit_count == 35\n    assert stats.reviewed_commit_count == 23\n    \n    ca = CodeReviewAnalysis(PYPI, "numpy", "1.21.4", "1.21.5")\n    stats = ca.stats\n    assert stats.phantom_files == 39\n    assert stats.files_with_phantom_lines == 1\n    assert stats.phantom_lines == 3\n    assert stats.reviewed_lines == 245\n    assert stats.non_reviewed_lines == 12\n    assert stats.total_commit_count == 10\n    assert stats.reviewed_commit_count == 9\n    \n    ca = CodeReviewAnalysis(NPM, "lodash", "4.17.20", "4.17.21")\n    stats = ca.stats\n    assert stats.phantom_files == 1046\n    assert stats.files_with_phantom_lines == 1\n    assert stats.phantom_lines == 1\n    assert stats.reviewed_lines == 58\n    assert stats.non_reviewed_lines == 14\n    assert stats.total_commit_count == 3\n    assert stats.reviewed_commit_count == 2\n    \n\nFuture Work\n------------\n\n* Also provides quality of the code review (e.g. what is the relation between the author and the reviewer? Was a sock account used to bypass code review check?)\n* What other checks should we perform for an update?\n\nCredits\n-------\n\nThis package was created with cookietemple_ using Cookiecutter_ based on Hypermodern_Python_Cookiecutter_.\n\n.. _cookietemple: https://cookietemple.com\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _PyPI: https://pypi.org/\n.. _Hypermodern_Python_Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _pip: https://pip.pypa.io/\n.. _Usage: https://depdive.readthedocs.io/en/latest/usage.html\n',
    'author': 'Nasif Imtiaz',
    'author_email': 'nasifimtiaz88@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nasifimtiazohi/depdive',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
